﻿using System;
using System.Collections.Generic;
using System.Linq;
using FluentNHibernate;
using FluentNHibernate.Automapping;
using FluentNHibernate.Automapping.Alterations;
using FluentNHibernate.Conventions;
using FluentNHibernate.Conventions.Instances;
using YesSql.Core.Indexes;

namespace YesSql.Core.Data.Mappings
{
    /// <summary>
    /// Adds a relationship to Document
    /// </summary>
    public class IndexAlteration : IAutoMappingAlteration
    {
        private readonly IEnumerable<Type> _hasDocumentsIndexes;

        public IndexAlteration(ITypeSource typeSource)
        {
            var types = typeSource.GetTypes().ToArray();
            _hasDocumentsIndexes = types.Where(x => typeof (IHasDocumentsIndex).IsAssignableFrom(x));
        }

        public void Alter(AutoPersistenceModel model)
        {
            var methods = model.GetType().GetMethods();
            var method = methods.Single(x => x.Name == "Override" && x.GetGenericArguments().Any());
            foreach (var type in _hasDocumentsIndexes)
            {
                // model.Override<{TIndex}>()
                var genericMethod = method.MakeGenericMethod(type);

                var alterationType = typeof (AlterationHelper<>).MakeGenericType(type);
                var overrideMethod = alterationType.GetMethods().Single(x => x.Name == "Override");
                var alteration = overrideMethod.Invoke(null, null);

                genericMethod.Invoke(model, new[] {alteration});
            }

            // automatically creates a database index on properties with the [Indexed] attribute
            model.Conventions.Setup(s => s.Add<IndexedPropertyConvention>());

            // configure the size of the column
            model.Conventions.Setup(s => s.Add<StringColumnLengthConvention>());
        }
    }

    internal class AlterationHelper<TIndex> where TIndex : IHasDocumentsIndex
    {
        public static Action<AutoMapping<TIndex>> Override()
        {
            return mapping => mapping.HasManyToMany(x => x.Documents)
                                  .AsSet()
                                  .Table("Documents_" + typeof (TIndex).Name)
                                  .Cascade.All()
                //.Inverse()
                ;

        }
    }

    public class IndexedPropertyConvention : AttributePropertyConvention<IndexedAttribute>
    {
        protected override void Apply(IndexedAttribute attribute, IPropertyInstance instance)
        {
            instance.Index("idx__" + instance.Property.Name);
        }
    }

    public class StringColumnLengthConvention : AttributePropertyConvention<StringLengthAttribute>
    {
        protected override void Apply(StringLengthAttribute attribute, IPropertyInstance instance)
        {
            instance.Length(attribute.Length);
        }
    }
}

